// desproxy-socks4server.c
//

#include "desproxy.h"
#define PROGRAM_NAME "desproxy-socks4server"
#define PROGRAM_VERSION VERSION

int request_sock,request_length;
int connstate[MAX_CONNECTIONS];
char remote_host[256],remote_port[6];
unsigned char nmethods[MAX_CONNECTIONS];
unsigned char methods[MAX_CONNECTIONS][256];

void EOC(int connection)
{
	debug_printf(">EOC(%d)\n",connection);
	debug_printf("connstate[%d]=%d\n",connection,connstate[connection]);
	client_sock_is_free[connection]=1;
	FD_CLR(client_sock[connection],&mask);
	FD_CLR(client_sock[connection],&rmask);
	close(client_sock[connection]);
	if (connstate[connection] == BICONNECTED)
	{
		debug_printf("connstate[connection]==BICONNECTED\n");
		FD_CLR(proxy_sock[connection],&mask);
		FD_CLR(proxy_sock[connection],&rmask);
		close(proxy_sock[connection]);
	}
	connstate[connection]=TO_RESET;
	debug_printf("EOC>\n");
}

void parse_command_line(int argc, char **argv)
{
	debug_printf(">parse_command_line(COMMAND_LINE)\n");
	if (argc !=4)
	{
		printf(gettext ("Usage: desproxy-socks4server proxy_host proxy_port local_port\n\n"));
        	exit(1);
        }
	
	proxy_host=argv[1];
	proxy_port=argv[2];
	local_port=argv[3];
	debug_printf("parse_command_line>\n");
}

void nothing_read(connection)
{
	debug_printf(">nothing_read()\n");
	if (read(client_sock[connection],buffer,1)==-1)
	{
		perror("read");
		EOC(connection);
		return;
	}
	if (buffer[0]!=4)
	{
		printf(gettext ("invalid client protocol version ("));
		printf("%d)\n",buffer[0]);
		strnsend(client_sock[connection],"\x00\x5b",2); //send general SOCKS error
		EOC(connection);
		return;
	}
	connstate[connection]=PROTO_V4_OK; //prococol version 4 accepted
	debug_printf("nothing_read>\n");
}

void proto_v4_ok(connection)
{
	debug_printf(">proto_v4_ok()\n");
	if (read(client_sock[connection],buffer,1)==-1)
	{
		perror("read");
		EOC(connection);
		return;
	}
	if (buffer[0]!=1) 
	{
		debug_printf("Method != 1 (CONNECT) => method refused\n");
		strnsend(client_sock[connection],"\x00\x5B\x00\x00\x00\x00\x00\x00",8);
		EOC(connection);
	} else {
		debug_printf("Method 1 (CONNECT) accepted\n");
		connstate[connection]=METHOD_ACCEPTED;
	}
	debug_printf("proto_v4_ok>\n");
	return;
}

void method_accepted(connection)
{
	debug_printf(">method_accepted()\n");
	if (read(client_sock[connection],buffer,6)==-1)
	{
		perror("read");
		EOC(connection);
		return;
	}
	debug_printf("IP ");
	debug_printf("(%d.%d.%d.%d:%d)\n",buffer[2],buffer[3],buffer[4],buffer[5],buffer[0]*256+buffer[1]);
	sprintf(remote_host,"%d.%d.%d.%d",buffer[2],buffer[3],buffer[4],buffer[5]);
	sprintf(remote_port,"%d",buffer[0]*256+buffer[1]);
	buffer[0]=0xff;
	while (buffer[0]!=0)
	{
		if (read(client_sock[connection],buffer,1)==-1)
		{
			perror("read");
			EOC(connection);
			return;
		}
	}
	if (connect_host_to_proxy(connection,remote_host,remote_port)!=0)
	{
		strnsend(client_sock[connection],"\x00\x5B\x00\x00\x00\x00\x00\x00",8);
		printf("connect_host_to_proxy: ERROR\n");
		EOC(connection);
		return;
	} else {
		strnsend(client_sock[connection],"\x00\x5A\x00\x00\x00\x00\x00\x00",8); //send sucess
		connstate[connection]=BICONNECTED;
	}
}

void biconnected(connection)
{
	int count;
	debug_printf(">biconnected()\n");
	debug_printf("client -> proxy (%d) ",connection);
	if ((count=read(client_sock[connection],buffer,sizeof(buffer)))==-1)
	{
		debug_printf("-1 bytes\n");
		perror("read");
		EOC(connection);
		return;
	}
	if (count==0) 
	{
		debug_printf("0 bytes\n");
		EOC(connection);
	} else { 
		if ((count=write(proxy_sock[connection],buffer,count))==-1)
		{
			perror("write");
			EOC(connection);
			return;
		}
		debug_printf("%d bytes\n",count);
	}
	debug_printf("biconnected>\n");
}

void process_connection_request(void)
{
	int connection;
	
	debug_printf(">process_connection_request()\n");
	for(connection=0;connection<MAX_CONNECTIONS;connection++) if (client_sock_is_free[connection]) break;
	if (connection == MAX_CONNECTIONS)
	{
		printf(gettext ("There are no connections available at this time\n"));
	} else {
		client_length=sizeof(client);
		client_sock[connection]=accept(request_sock,(struct sockaddr *)&client,&client_length);
		if (client_sock[connection] < 0)
		{
			perror("accept");
			EOC(connection);
			return;
		}
		debug_printf("accept> (%d)\n",client_sock[connection]);
		client_sock_is_free[connection]=0;
		printf(gettext ("Connection request from %s, port %d\n"),inet_ntoa(client.sin_addr),ntohs(client.sin_port));
		if (client_sock[connection] >maxfd) maxfd=client_sock[connection];
		FD_SET(client_sock[connection],&mask);
		connstate[connection]=NOTHING_READ;
	}
	debug_printf("process_connection_request>\n");
}

int main (int argc, char **argv)
{
	int connection;
	int count;
	int nfound,nfound2;
	struct timeval timeout;
	
	signal(SIGPIPE,(sig_t)PIPE);
	initialize_gettext();
	print_program_version(PROGRAM_NAME,PROGRAM_VERSION);
	parse_command_line(argc,argv);
	
	request_sock=listen_in_TCP_port(atoi(local_port));
	printf(gettext ("Press ENTER to Quit\n"));
	mark_all_client_socks_as_free();
	
	FD_ZERO(&mask);
	FD_SET(request_sock,&mask);
	FD_SET(0,&mask);
	maxfd=request_sock;
	while (1)
	{
		rmask=mask;
		timeout.tv_sec=5;
		timeout.tv_usec=0;
		nfound=select(maxfd+1,&rmask,NULL,NULL, &timeout);
		if (DEBUG>1) debug_printf("nfound> (%d)\n",nfound);
		if (nfound<0)
		{
			if (errno==EINTR)
			{
				printf(gettext ("Interrupted by system call\n"));
				continue;
			}
			perror("select");
			exit(1);
		}
		
		nfound2=nfound; // TEST!
		
		if (FD_ISSET(0,&rmask)) // if FD_ISSET(stdin) (Carriage Return, ENTER) then exit
		{
			close(request_sock);
			exit(0);
		}
		
		if (FD_ISSET(request_sock, &rmask)) // processes connection requests
		{
			nfound2--;
			process_connection_request();
		}

		for (connection=0;connection<MAX_CONNECTIONS;connection++) // incoming connections 
		{
			if (DEBUG>1) debug_printf("connection #%d\n",connection);
			if (!client_sock_is_free[connection])
			{
				debug_printf("client_sock_is_free[%d]=%d\n",connection,client_sock_is_free[connection]);
				if (FD_ISSET(client_sock[connection],&rmask))
				{
					nfound2--;
					if (DEBUG>1) debug_printf("FD_ISSET(client_sock[%d],&rmask\n",connection);
					if (connstate[connection]==BICONNECTED)
					{
						biconnected(connection);
					}
					if (connstate[connection]==METHOD_ACCEPTED)
					{
						method_accepted(connection);
					}
					if (connstate[connection]==PROTO_V4_OK)
					{
						proto_v4_ok(connection);
					}
					if (connstate[connection]==NOTHING_READ)
					{
						nothing_read(connection);
					}
				}
				if (connstate[connection]==BICONNECTED)
				{
					if (FD_ISSET(proxy_sock[connection],&rmask))
					{
						nfound2--;
						debug_printf("proxy -> client (%d) ",connection);
						if ((count=read(proxy_sock[connection],buffer,sizeof(buffer)))==-1)
						{
							debug_printf("-1 bytes\n");
							perror("read");
							EOC(connection);
						} else {
							if (count==0)
							{
								debug_printf("0 bytes\n");
								EOC(connection);
							} else {
								if ((count=write(client_sock[connection],buffer,count))==-1)
								{
									perror("write");
									EOC(connection);
								}
								debug_printf("%d bytes\n",count);
							}
						}
					}
				}
			}
		}

		if (DEBUG>1)
		{
			printf("nfound %d;nfound2 %d\n",nfound,nfound2);
			for(connection=0;connection<MAX_CONNECTIONS;connection++)
			{
				printf("connstate[%d]=%d\n",connection,connstate[connection]);
				if (FD_ISSET(proxy_sock[connection],&rmask))
				{
					printf("FD_ISSET(proxy_sock[%d],&rmask)\n",connection);
				}
				if (FD_ISSET(client_sock[connection],&rmask))
				{
					printf("FD_ISSET(client_sock[%d],&rmask)\n",connection);
				}
			}
			fflush(stdout);
			if (nfound2!=0)
			{
//				exit(0);
			}
		}
	}
	exit(0);
}
