// desproxy-socks5server.c
//

#include "desproxy.h"
#define PROGRAM_NAME "desproxy-socks5server"
#define PROGRAM_VERSION VERSION

int request_sock,request_length;
int connstate[MAX_CONNECTIONS];
char remote_host[256],remote_port[6];
unsigned char nmethods[MAX_CONNECTIONS];
unsigned char methods[MAX_CONNECTIONS][256];

/*
 * Terminates connections gracefully
 */
void EOC(int connection)
{
	debug_printf(">EOC(%d)\n",connection);
	debug_printf("connstate[%d]=%d\n",connection,connstate[connection]);
	client_sock_is_free[connection]=1;
	FD_CLR(client_sock[connection],&mask);
	FD_CLR(client_sock[connection],&rmask);
	close(client_sock[connection]);
	if (connstate[connection] == BICONNECTED)
	{
		debug_printf("connstate[connection]==BICONNECTED\n");
		FD_CLR(proxy_sock[connection],&mask);
		FD_CLR(proxy_sock[connection],&rmask);
		close(proxy_sock[connection]);
	}
	connstate[connection]=TO_RESET;
	debug_printf("EOC>\n");
}

/*
 * Parses command line
 */
void parse_command_line(int argc, char **argv)
{
	debug_printf(">parse_command_line(COMMAND_LINE)\n");
	if (argc !=4)
	{
		printf(gettext ("Usage: desproxy-socks5server proxy_host proxy_port local_port\n\n"));
        	exit(1);
        }
	
	proxy_host=argv[1];
	proxy_port=argv[2];
	local_port=argv[3];
	debug_printf("parse_command_line>\n");
}

/*
 * invoked for the first time something is read
 */
void nothing_read(connection)
{
	debug_printf(">nothing_read()\n");
	if (read(client_sock[connection],buffer,1)==-1)
	{
		perror("read");
		EOC(connection);
		return;
	}
	if (buffer[0]!=5)
	{
		printf(gettext ("invalid client protocol version ("));
		printf("%d)\n",buffer[0]);
		strnsend(client_sock[connection],"\x05\x01\x00\x00",4); //send general SOCKS error
		EOC(connection);
		return;
	}
	connstate[connection]=PROTO_V5_OK; //prococol version 5 accepted
	debug_printf("nothing_read>\n");
}

/*
 * invoked when we read the protocol header and it is 5
 */
void proto_v5_ok(connection)
{
	debug_printf(">proto_v5_ok()\n");
	if (read(client_sock[connection],buffer,1)==-1)
	{
		perror("read");
		EOC(connection);
		return;
	}
	nmethods[connection]=buffer[0];
	connstate[connection]=NMETHODS_READ; //NMETHODS is read
	debug_printf("Number of methods: %d\n",nmethods[connection]);
	debug_printf("proto_v5_ok>\n");
	return;
}

/*
 * Invoked when the method is accepted
 */
void method_accepted(connection)
{
	unsigned char index;
	
	debug_printf(">method_accepted()\n");
	if (read(client_sock[connection],buffer,4)==-1)
	{
		perror("read");
		EOC(connection);
		return;
	}
	debug_printf("VER %d\n",buffer[0]);
	debug_printf("CMD %d\n",buffer[1]);
	debug_printf("RSV %d\n",buffer[2]);
	debug_printf("ATYP %d\n",buffer[3]);
	if (buffer[1]!=1)
	{
		debug_printf("Command %d not supported\n",buffer[1]);
		/*
		 * Send "Command not supported"
		 */
		strnsend(client_sock[connection],"\x05\x07\x00\x01",4);
		strnsend(client_sock[connection],(char *)&server.sin_port,2);
		strnsend(client_sock[connection],(char *)&server.sin_addr.s_addr,4);
		EOC(connection);
		return;
	}
	if (buffer[3]==1)
	{
		if(read(client_sock[connection],buffer,6)==-1)
		{
			perror("read");
			EOC(connection);
			return;
		}
		debug_printf("ATYP = IP ");
		debug_printf("(%d.%d.%d.%d:%d)\n",buffer[0],buffer[1],buffer[2],buffer[3],buffer[4]*256+buffer[5]);
		sprintf(remote_host,"%d.%d.%d.%d",buffer[0],buffer[1],buffer[2],buffer[3]);
		sprintf(remote_port,"%d",buffer[4]*256+buffer[5]);
	}
	if (buffer[3]==3)
	{
		if (read(client_sock[connection],buffer,1)==-1)
		{
			perror("read");
			EOC(connection);
			return;
		}
		debug_printf("ATYP = DOMAINNAME\n");
		strcpy(remote_host,"");
		for (index=buffer[0];index>0;index--)
		{
			if (read(client_sock[connection],buffer,1)==-1)
			{
				perror("read");
				EOC(connection);
				return;
			}
			strncat(remote_host,buffer,1);
		}
		strncat(remote_host,"\x00",1);
		debug_printf("remote_host %s\n",remote_host);
		if (read(client_sock[connection],buffer,2)==-1)
		{
			perror("read");
			EOC(connection);
			return;
		}
		sprintf(remote_port,"%d",buffer[0]*256+buffer[1]);
		debug_printf("remote_port %s\n",remote_port);
	}
	if (connect_host_to_proxy(connection,remote_host,remote_port)!=0)
	{
		/*
		 * Send "Host unreachable"
		 */
		strnsend(client_sock[connection],"\x05\x04\x00\x01",4); 
		printf("connect_host_to_proxy: ERROR\n");
		EOC(connection);
		return;
	} else {
		/*
		 * Send "Success"
		 */
		strnsend(client_sock[connection],"\x05\x00\x00\x01",4); 
		strnsend(client_sock[connection],(char *)&server.sin_port,2);
		strnsend(client_sock[connection],(char *)&server.sin_addr.s_addr,4);
		connstate[connection]=BICONNECTED;
	}
}

void nmethods_read(connection)
{
	unsigned char index;
	debug_printf(">nmethods_read()\n");
	if (read(client_sock[connection],&methods[connection][0],nmethods[connection])==-1)
	{
		perror("read");
		EOC(connection);
		return;
	}
	connstate[connection]=LOOKING_FOR_METHODS;
	for (index=0;index<nmethods[connection];index++)
	{
		/*
		 * if there is one method 0 (NOAUTH) then accept it
		 */
		if (methods[connection][index] ==0)
		{
			connstate[connection]=METHOD_ACCEPTED;
			strnsend(client_sock[connection],"\x05\x00",2);
			debug_printf("Selected method 0\n");
		}
	}
	if (connstate[connection]!=METHOD_ACCEPTED)
	{
		debug_printf(gettext ("All methods rejected\n"));
		strnsend(client_sock[connection],"\x05\xff",2);
		EOC(connection);
		return;
	}
	debug_printf("nmethods_read>\n");
}

/*
 * Invoked for transcactions when the client is accepted yet
 */
void biconnected(connection)
{
	int count;
	debug_printf(">biconnected()\n");
	debug_printf("client -> proxy (%d) ",connection);
	if ((count=read(client_sock[connection],buffer,sizeof(buffer)))==-1)
	{
		debug_printf("-1 bytes\n");
		perror("read");
		EOC(connection);
		return;
	}
	if (count==0) 
	{
		debug_printf("0 bytes\n");
		EOC(connection);
	} else { 
		if ((count=write(proxy_sock[connection],buffer,count))==-1)
		{
			perror("write");
			EOC(connection);
			return;
		}
		debug_printf("%d bytes\n",count);
	}
	debug_printf("biconnected>\n");
}

/*
 * Processes connection requests
 */
void process_connection_request(void)
{
	int connection;
	
	debug_printf(">process_connection_request()\n");
	for(connection=0;connection<MAX_CONNECTIONS;connection++) if (client_sock_is_free[connection]) break;
	if (connection == MAX_CONNECTIONS)
	{
		printf(gettext ("There are no connections available at this time\n"));
	} else {
		client_length=sizeof(client);
		client_sock[connection]=accept(request_sock,(struct sockaddr *)&client,&client_length);
		if (client_sock[connection] < 0)
		{
			perror("accept");
			EOC(connection);
			return;
		}
		debug_printf("accept> (%d)\n",client_sock[connection]);
		client_sock_is_free[connection]=0;
		printf(gettext ("Connection request from %s, port %d\n"),inet_ntoa(client.sin_addr),ntohs(client.sin_port));
		if (client_sock[connection] >maxfd) maxfd=client_sock[connection];
		FD_SET(client_sock[connection],&mask);
		connstate[connection]=NOTHING_READ;
	}
	debug_printf("process_connection_request>\n");
}

int main (int argc, char **argv)
{
	int connection;
	int count;
	int nfound,nfound2;
	struct timeval timeout;
	
	signal(SIGPIPE,(sig_t)PIPE);
	initialize_gettext();
	print_program_version(PROGRAM_NAME,PROGRAM_VERSION);
	parse_command_line(argc,argv);
	
	request_sock=listen_in_TCP_port(atoi(local_port));
	printf(gettext ("Press ENTER to Quit\n"));
	mark_all_client_socks_as_free();
	
	FD_ZERO(&mask);
	FD_SET(request_sock,&mask);
	FD_SET(0,&mask);
	maxfd=request_sock;
	while (1)
	{
		rmask=mask;
		timeout.tv_sec=5;
		timeout.tv_usec=0;
		nfound=select(maxfd+1,&rmask,NULL,NULL, &timeout);
		debug_printf("nfound> (%d)\n",nfound);
		if (nfound<0)
		{
			if (errno==EINTR)
			{
				printf(gettext ("Interrupted by system call\n"));
				continue;
			}
			perror("select");
			exit(1);
		}
		
		nfound2=nfound; // TEST!
		
		if (FD_ISSET(0,&rmask)) // if FD_ISSET(stdin) (Carriage Return, ENTER) then exit
		{
			close(request_sock);
			exit(0);
		}
		
		if (FD_ISSET(request_sock, &rmask)) // processes connection requests
		{
			nfound2--;
			process_connection_request();
		}

		for (connection=0;connection<MAX_CONNECTIONS;connection++) // incoming connections 
		{
			if (DEBUG>1) debug_printf("connection #%d\n",connection);
			
			if (!client_sock_is_free[connection])
			{
				debug_printf("client_sock_is_free[%d]=%d\n",connection,client_sock_is_free[connection]);
				if (FD_ISSET(client_sock[connection],&rmask))
				{
					nfound2--;
					debug_printf("FD_ISSET(client_sock[%d],&rmask\n",connection);
					if (connstate[connection]==BICONNECTED)
					{
						biconnected(connection);
					}
					if (connstate[connection]==METHOD_ACCEPTED)
					{
						method_accepted(connection);
					}
					if (connstate[connection]==NMETHODS_READ)
					{
						nmethods_read(connection);
					}
					if (connstate[connection]==PROTO_V5_OK)
					{
						proto_v5_ok(connection);
					}
					if (connstate[connection]==NOTHING_READ)
					{
						nothing_read(connection);
					}
				}
				if (connstate[connection]==BICONNECTED)
				{
					if (FD_ISSET(proxy_sock[connection],&rmask))
					{
						nfound2--;
						debug_printf("proxy -> client (%d) ",connection);
						if ((count=read(proxy_sock[connection],buffer,sizeof(buffer)))==-1)
						{
							perror("read");
							EOC(connection);
						} else {
							if (count==0)
							{
								close(client_sock[connection]);
								close(proxy_sock[connection]);
								FD_CLR(client_sock[connection], &mask);
								FD_CLR(proxy_sock[connection], &mask);
								client_sock_is_free[connection]=1;
								printf(gettext ("End of connection\n"));
							} else {
								if ((count=write(client_sock[connection],buffer,count))==-1)
								{
									perror("write");
									EOC(connection);
								}
							}
							debug_printf("%d bytes\n",count);
						}
					}
				}
			}
		}

		if (DEBUG>1)
		{
			printf("nfound %d;nfound2 %d\n",nfound,nfound2);
			for(connection=0;connection<MAX_CONNECTIONS;connection++)
			{
				printf("connstate[%d]=%d\n",connection,connstate[connection]);
				if (FD_ISSET(proxy_sock[connection],&rmask))
				{
					printf("FD_ISSET(proxy_sock[%d],&rmask)\n",connection);
				}
				if (FD_ISSET(client_sock[connection],&rmask))
				{
					printf("FD_ISSET(client_sock[%d],&rmask)\n",connection);
				}
			}
			fflush(stdout);
			if (nfound2!=0)
			{
// 				exit(0);
			}
		}
	}
	exit(0);
}
